const App = @import("./App.zig");

pub const Queue = @import("./gfx/Queue.zig");

const coral = @import("coral");

const Device = @import("./gfx/Device.zig");

const ext = @import("./ext.zig");

const msg = @import("./msg.zig");

const std = @import("std");

pub const Color = @Vector(4, f32);

pub const Display = struct {
	sdl_window: *ext.SDL_Window,
	device: Device,

	pub fn resize(self: Display, width: u16, height: u16) void {
		ext.SDL_SetWindowSize(self.sdl_window, width, height);
	}

	pub fn retitle(self: Display, title: []const u8) void {
		var sentineled_title = [_:0]u8{0} ** 255;

		@memcpy(sentineled_title[0 .. @min(title.len, sentineled_title.len)], title);
		ext.SDL_SetWindowTitle(self.sdl_window, &sentineled_title);
	}

	pub fn set_resizable(self: Display, resizable: bool) void {
		ext.SDL_SetWindowResizable(self.sdl_window, @intFromBool(resizable));
	}
};

pub const Handle = enum (usize) {
	none,
	_,

	pub fn index(self: Handle) ?usize {
		return switch (self) {
			.none => null,
			_ => @intFromEnum(self) - 1,
		};
	}
};

pub const Input = union (enum) {
	key_up: Key,
	key_down: Key,

	pub const Key = enum (u32) {
		a = ext.SDL_SCANCODE_A,
		d = ext.SDL_SCANCODE_D,
		s = ext.SDL_SCANCODE_S,
		w = ext.SDL_SCANCODE_W,

		pub fn scancode(key: Key) u32 {
			return @intFromEnum(key);
		}
	};
};

pub const Point2D = @Vector(2, f32);

pub const Transform2D = extern struct {
	xbasis: Point2D = .{1, 0},
	ybasis: Point2D = .{0, 1},
	origin: Point2D = @splat(0),
};

pub const Work = struct {
	queue: *Queue.Buffer,
	primitives: *const Primitives,

	const Primitives = struct {
		quad_mesh: Handle,
	};

	pub const State = struct {
		queue: *Queue,
		primitives: *const Primitives,
	};

	pub fn bind(context: coral.system.BindContext) std.mem.Allocator.Error!State {
		const queue = try Queue.create();

		return .{
			.primitives = (try context.register_read_only_resource_access(.none, Primitives)) orelse create: {
				const buffer = queue.pending();
				const half_extent = 0.5;

				try context.world.set_resource(.none, Primitives{
					.quad_mesh = try buffer.open(.{
						.label = "quad mesh primitive",

						.resource = .{
							.mesh_2d = .{
								.indices = &.{0, 1, 2, 0, 2, 3},

								.vertices = &.{
									.{.xy = .{-half_extent, half_extent}, .uv = .{0, 1}},
									.{.xy = .{half_extent, half_extent}, .uv = .{1, 1}},
									.{.xy = .{half_extent, -half_extent}, .uv = .{1, 0}},
									.{.xy = .{-half_extent, -half_extent},  .uv = .{0, 0}},
								},
							},
						},
					}),
				});

				break: create (try context.register_read_only_resource_access(.none, Primitives)).?;
			},

			.queue = queue,
		};
	}

	pub fn init(state: *State) Work {
		return .{
			.queue = state.queue.pending(),
			.primitives = state.primitives,
		};
	}

	pub fn unbind(state: *State) void {
		state.queue.release();
	}
};

pub const colors = struct {
	pub const black = greyscale(0);

	pub const white = greyscale(1);

	pub fn greyscale(v: f32) Color {
		return .{v, v, v, 1};
	}

	pub fn rgb(r: f32, g: f32, b: f32) Color {
		return .{r, g, b, 1};
	}
};

pub fn poll(app: coral.Write(App), inputs: msg.Send(Input)) !void {
	var event = @as(ext.SDL_Event, undefined);

	while (ext.SDL_PollEvent(&event) != 0) {
		switch (event.type) {
			ext.SDL_QUIT => app.res.quit(),
			ext.SDL_KEYUP => try inputs.push(.{.key_up = @enumFromInt(event.key.keysym.scancode)}),
			ext.SDL_KEYDOWN => try inputs.push(.{.key_down = @enumFromInt(event.key.keysym.scancode)}),
			else => {},
		}
	}
}

pub fn setup(world: *coral.World, events: App.Events) (error {SDLError} || std.Thread.SpawnError || std.mem.Allocator.Error)!void {
	if (ext.SDL_Init(ext.SDL_INIT_VIDEO) != 0) {
		return error.SDLError;
	}

	const sdl_window = create: {
		const position = ext.SDL_WINDOWPOS_CENTERED;
		const flags = ext.SDL_WINDOW_OPENGL;
		const width = 640;
		const height = 480;

		break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
			return error.SDLError;
		};
	};

	errdefer ext.SDL_DestroyWindow(sdl_window);

	var device = try Device.init(sdl_window);

	errdefer device.deinit();

	try world.set_resource(.main, Display{
		.device = device,
		.sdl_window = sdl_window,
	});

	try world.on_event(events.pre_update, coral.system_fn(poll), .{.label = "poll gfx"});
	try world.on_event(events.exit, coral.system_fn(stop), .{.label = "stop gfx"});
	try world.on_event(events.finish, coral.system_fn(submit), .{.label = "submit gfx"});
}

pub fn stop(display: coral.WriteBlocking(Display)) void {
	display.res.device.deinit();
	ext.SDL_DestroyWindow(display.res.sdl_window);
}

pub fn submit(display: coral.WriteBlocking(Display)) void {
	display.res.device.submit();
}
